""" Tools for Git versioned projects. """

from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Literal

from kraken.core import Project

from kraken.std.util.check_file_contents_task import CheckFileContentsTask

from . import tasks
from .config import dump_gitconfig, load_gitconfig
from .version import GitVersion, git_describe

__all__ = [
    "tasks",
    "load_gitconfig",
    "dump_gitconfig",
    "git_describe",
    "GitVersion",
    "gitignore",
]


def gitignore(
    *,
    name: str = "gitignore",
    group: str = "apply",
    check_group: str = "check",
    gitignore_file: str | Path = ".gitignore",
    generated_content: Sequence[str] | None = (),
    gitignore_io_tokens: Sequence[str] = (),
    gitignore_io_allow_http_request_backfill: bool = False,
    where: Literal["top", "bottom"] = "top",
    project: Project | None = None,
) -> tuple[tasks.GitignoreSyncTask, CheckFileContentsTask]:
    """
    Creates a #GitignoreSyncTask and #CheckFileContentsTask for the given project.
    """

    project = project or Project.current()
    task = project.task(name, tasks.GitignoreSyncTask, group=group)
    task.file.set(Path(gitignore_file))
    if generated_content is not None:
        task.generated_content.setmap(lambda x: [*x, *generated_content])  # type: ignore[misc]  # See https://github.com/python/mypy/issues/14891  # noqa: E501
    task.gitignore_io_tokens.set(list(gitignore_io_tokens))
    task.gitignore_io_allow_http_request_backfill.set(gitignore_io_allow_http_request_backfill)
    task.where.set(where)
    return task, task.create_check(group=check_group)



"""
Manages a `mitmweb` instance to run in the background for injecting basic-auth into requests to hosts for which
credentials are passed down. The `mitmweb` web interface can be reached on `localhost:8900` while it is running.
The proxy stays alive as a daemon process until its configuration changes.
"""

from __future__ import annotations

import json
import logging
import time
from collections.abc import Mapping, Sequence
from pathlib import Path

from kraken.std.util.daemon_controller import DaemonController

logger = logging.getLogger(__name__)

daemon_state_file = Path("~/.config/krakenw/.mitmweb-daemon-state.json").expanduser()
daemon_log_file = daemon_state_file.with_suffix(".log")
mitmweb_port = 8899
mitmweb_ui_port = 8900
mitmproxy_ca_cert_file = Path("~/.mitmproxy/mitmproxy-ca-cert.pem").expanduser()
inject_auth_addon_file = Path(__file__).parent / "mitm_addon.py"


def start_mitmweb_proxy(
    auth: Mapping[str, tuple[str, str]], startup_wait_time: float = 3.0, additional_args: Sequence[str] = ()
) -> tuple[str, Path]:
    controller = DaemonController("kraken.mitmweb", daemon_state_file)
    started = controller.run(
        command=[
            "mitmweb",
            "--no-web-open-browser",
            "--web-port",
            str(mitmweb_ui_port),
            "--listen-port",
            str(mitmweb_port),
            "-s",
            str(inject_auth_addon_file),
            "--set",
            "auth=" + json.dumps(auth),
            # `mitmproxy` buffers the entire response before forwarding it to
            # the client. This is problematic when e.g. cloning large git repos
            # via http. As we're not using any filtering mechanism, we can just
            # stream the bodies through without `mitmproxy` storing them.
            #
            # See https://github.com/mitmproxy/mitmproxy/issues/6237 for
            # context.
            "--set",
            "stream_large_bodies=3m",
            *additional_args,
        ],
        cwd=Path("~").expanduser(),
        stdout=daemon_log_file,
        stderr="stdout",
    )
    if started:
        # Give the proxy some time to fully start up.
        time.sleep(startup_wait_time)
        if not controller.is_alive():
            raise RuntimeError("The mitmweb proxy failed to start. Check its logs at %s" % daemon_log_file)
    return f"localhost:{mitmweb_port}", mitmproxy_ca_cert_file


def stop_mitmweb_proxy() -> None:
    controller = DaemonController("kraken.mitmweb", daemon_state_file)
    controller.stop()



""" Implements PDM as a build system for kraken-std. """

from __future__ import annotations

import logging
import os
import shutil
import subprocess as sp
from collections.abc import Sequence
from pathlib import Path
from typing import Any

from kraken.common import NotSet
from kraken.common.path import is_relative_to
from kraken.core import TaskStatus

from kraken.std.python.pyproject import PackageIndex, Pyproject, PyprojectHandler
from kraken.std.python.settings import PythonSettings

from . import ManagedEnvironment, PythonBuildSystem

logger = logging.getLogger(__name__)


class PdmPyprojectHandler(PyprojectHandler):
    """
    Implements the PyprojectHandler interface for PDM projects.
    """

    def __init__(self, pyproj: Pyproject) -> None:
        super().__init__(pyproj)

    # PyprojectHandler

    def get_package_indexes(self) -> list[PackageIndex]:
        results = []
        for source in self.raw.get("tool", {}).get("pdm", {}).get("source", []):
            if "username" in source or "password" in source:
                logger.warning(
                    "username/password in pyproject.toml [tool.pdm.source] is not supported, the information may "
                    "be lost (source name: %s)",
                    source["name"],
                )
            if source.get("type") == "find_links":
                logger.warning(
                    "source of type 'find_links' in pyproject.toml [tool.pdm.source] is not supported "
                    "(source name: %s)",
                    source["name"],
                )
                continue
            # TODO: To understand if a source is a "default" source as per Poetry's definition, we need
            #       to know if `pypi.ignore_stored_index` is set to `true`. It appears that this may be a
            #       local (for the project) or global (for the user) option. For now, we assume that this
            #       option is not turned on (as it is not by default) and all sources we find are primary
            #       sources instead of default sources. https://pdm.fming.dev/latest/usage/config/
            # TODO: We may also want to consider ensuring that respect-source-order is enabled, see
            #       https://pdm.fming.dev/latest/usage/config/#respect-the-order-of-the-sources
            results.append(
                PackageIndex(
                    alias=source["name"],
                    index_url=source["url"],
                    verify_ssl=source.get("verify_ssl", True),
                    priority=PackageIndex.Priority.primary,
                )
            )
        return results

    def set_package_indexes(self, indexes: Sequence[PackageIndex]) -> None:
        """
        Set the list of package indexes to the given list of indexes, replacing any existing indexes.

        Note that this only updates the package indexes in the `[tool.pdm.source]` section of the
        pyproject.toml file.
        See https://pdm.fming.dev/latest/usage/config/#source for more details.
        """

        sources_conf = self.raw.setdefault("tool", {}).setdefault("pdm", {}).setdefault("source", [])
        sources_conf.clear()

        # We don't currently support fully replicating the semantics of the index priority in PDM.
        # Instead, we treat all as primaries but add secondary/supplement to the end of the list.
        if any(x.priority == PackageIndex.Priority.default for x in indexes):
            logger.warning(
                "default index priority is not supported in %s, treating them as primary instead and "
                "putting them in front",
                type(self).__name__,
            )
        if any(x.priority in (PackageIndex.Priority.secondary, PackageIndex.Priority.supplemental) for x in indexes):
            logger.warning(
                "secondary/supplement index priority is not supported in %s, treating them as primary instead and "
                "putting them at the back",
                type(self).__name__,
            )
        indexes = sorted(
            indexes,
            key=lambda x: 0
            if x.priority == PackageIndex.Priority.default
            else 1
            if x.priority == PackageIndex.Priority.primary
            else 2,
        )

        for index in indexes:
            source: dict[str, Any] = {"name": index.alias, "url": index.index_url}
            if index.verify_ssl is False:
                source["verify_ssl"] = False
            sources_conf.append(source)


class PDMPythonBuildSystem(PythonBuildSystem):
    name = "PDM"

    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory

    def get_pyproject_reader(self, pyproject: Pyproject) -> PdmPyprojectHandler:
        return PdmPyprojectHandler(pyproject)

    def supports_managed_environments(self) -> bool:
        return True

    def get_managed_environment(self) -> ManagedEnvironment:
        return PDMManagedEnvironment(self.project_directory)

    def update_lockfile(self, settings: PythonSettings, pyproject: Pyproject) -> TaskStatus:
        command = ["pdm", "update"]
        sp.check_call(command, cwd=self.project_directory)
        return TaskStatus.succeeded()

    def requires_login(self) -> bool:
        return True

    def login(self, settings: PythonSettings) -> None:
        for index in settings.package_indexes.values():
            if index.is_package_source and index.credentials:
                commands = [
                    ["pdm", "config", f"pypi.{index.alias}.url", index.index_url],
                    [
                        "pdm",
                        "config",
                        f"pypi.{index.alias}.username",
                        index.credentials[0],
                    ],
                    [
                        "pdm",
                        "config",
                        f"pypi.{index.alias}.password",
                        index.credentials[1],
                    ],
                ]
                for command in commands:
                    safe_command = command[:-1] + ["MASKED"]
                    logger.info("$ %s", safe_command)

                    code = sp.call(command)
                    if code != 0:
                        raise RuntimeError(f"command {safe_command!r} failed with exit code {code}")

    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        previous_version: str | None = None

        if as_version is not None:
            # Bump the in-source version number.
            pyproject = self.get_pyproject_reader(Pyproject.read(self.project_directory / "pyproject.toml"))
            previous_version = pyproject.get_version()
            # pyproject.set_path_dependencies_to_version(as_version)
            pyproject.set_version(as_version)
            pyproject.raw.save()

        # PDM does not allow configuring the output folder, so it's always going to be "dist/".
        # We remove the contents of that folder to make sure we know what was produced.
        dist_dir = self.project_directory / "dist"
        if dist_dir.exists():
            shutil.rmtree(dist_dir)

        command = ["pdm", "build"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)

        src_files = list(dist_dir.iterdir())
        dst_files = [output_directory / path.name for path in src_files]
        os.makedirs(output_directory, exist_ok=True)
        for src, dst in zip(src_files, dst_files):
            shutil.move(str(src), dst)

        # Unless the output directory is a subdirectory of the dist_dir, we remove the dist dir again.
        if not is_relative_to(output_directory, dist_dir):
            shutil.rmtree(dist_dir)

        # Roll back the previously updated in-source version numbers.
        if previous_version is not None:
            pyproject.set_version(previous_version)
            pyproject.raw.save()

        return dst_files


class PDMManagedEnvironment(ManagedEnvironment):
    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory
        self._env_path: Path | None | NotSet = NotSet.Value

    def _get_pdm_environment_path(self, create: bool) -> None | Path:
        """Uses `pdm venv --path in-project`. TODO(simone.zandara) Add support for more environments."""

        get_command = ["pdm", "venv", "--path", "in-project"]
        logger.debug("$ %s", get_command)
        try:
            return Path(sp.check_output(get_command, cwd=self.project_directory, stderr=sp.PIPE).decode().strip())
        except sp.CalledProcessError as exc:
            logger.debug("pdm venv --path failed with exit code %s, output: %s", exc.returncode, exc.stderr.decode())
            if not create:
                return None

        create_command = ["pdm", "venv", "create"]
        logger.info("$ %s", create_command)
        sp.check_call(create_command, cwd=self.project_directory)

        # Make sure we use the in-project environment.
        use_command = ["pdm", "use", "--venv", "in-project"]
        logger.info("$ %s", use_command)
        sp.check_call(use_command, cwd=self.project_directory)

        path = self._get_pdm_environment_path(create=False)
        if path is None:
            raise RuntimeError("Failed to create PDM environment")

        return path

    # ManagedEnvironment

    def exists(self) -> bool:
        try:
            self.get_path()
            return True
        except RuntimeError:
            return False

    def get_path(self) -> Path:
        if self._env_path is NotSet.Value:
            self._env_path = self._get_pdm_environment_path(create=False)
        if self._env_path is None:
            raise RuntimeError("managed environment does not exist")
        return self._env_path

    def install(self, settings: PythonSettings) -> None:
        self._get_pdm_environment_path(create=True)

        command = ["pdm", "install"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)



from __future__ import annotations

from pathlib import Path

from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class MypyStubtestTask(EnvironmentAwareDispatchTask):
    description = "Static validation for Python type stubs using Mypy."
    python_dependencies = ["mypy"]

    package: Property[str]
    ignore_missing_stubs: Property[bool] = Property.default(False)
    ignore_positional_only: Property[bool] = Property.default(False)
    allowlist: Property[Path]
    mypy_config_file: Property[Path]

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        command = ["python", "-m", "mypy.stubtest", self.package.get()]
        if self.ignore_missing_stubs.get():
            command.append("--ignore-missing-stub")
        if self.ignore_positional_only.get():
            command.append("--ignore-positional-only")
        if self.allowlist.is_filled():
            command.extend(("--allowlist", str(self.allowlist.get().absolute())))
        if self.mypy_config_file.is_filled():
            command.extend(("--mypy-config-file", str(self.mypy_config_file.get().absolute())))
        return command


def mypy_stubtest(
    *,
    name: str = "python.mypy.stubtest",
    project: Project | None = None,
    package: str,
    ignore_missing_stubs: bool = False,
    ignore_positional_only: bool = False,
    allowlist: Path | None = None,
    mypy_config_file: Path | None = None,
) -> MypyStubtestTask:
    project = project or Project.current()
    task = project.task(name, MypyStubtestTask, group="lint")
    task.package = package
    task.ignore_missing_stubs = ignore_missing_stubs
    task.ignore_positional_only = ignore_positional_only
    task.allowlist = allowlist
    task.mypy_config_file = mypy_config_file
    return task



import pytest

from kraken.std.git.gitignore.gitignore_io import TOKENS, gitignore_io_fetch_cached, load_token_cache


def test__gitignore_io_cache_for_every_token() -> None:
    assert set(load_token_cache()) == set(TOKENS)


def test__gitignore_io_fetch_cached__errors_without_backfil_on_missing_token() -> None:
    with pytest.raises(ValueError) as excinfo:
        gitignore_io_fetch_cached(["missing-token"], backfill=False)
    assert str(excinfo.value) == (
        "The following gitignore.io tokens are not distributed as part of kraken-std: missing-token\n"
        "Backfill is disabled, so this error is raised instead of making another HTTP request to gitignore.io."
    )



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from kraken.common import import_class
from kraken.core import Project, Task

from kraken.std.docker.tasks.base_build_task import BaseBuildTask
from kraken.std.docker.tasks.manifest_tool_push_task import ManifestToolPushTask

__all__ = ["build_docker_image", "manifest_tool", "sidecar_container"]

DEFAULT_BUILD_BACKEND = "native"
BUILD_BACKENDS = {
    "buildx": f"{__name__}.tasks.buildx_build_task.BuildxBuildTask",
    "kaniko": f"{__name__}.tasks.kaniko_build_task.KanikoBuildTask",
    "native": f"{__name__}.tasks.docker_build_task.DockerBuildTask",
}


def build_docker_image(
    *,
    name: str = "buildDocker",
    backend: str = DEFAULT_BUILD_BACKEND,
    project: Project | None = None,
    **kwds: Any,
) -> BaseBuildTask:
    """Create a new task in the current project that builds a Docker image and eventually pushes it."""

    task_class = import_class(BUILD_BACKENDS[backend], BaseBuildTask)  # type: ignore[type-abstract]
    return (project or Project.current()).do(name, task_class, **kwds)


def manifest_tool(
    *,
    name: str,
    template: str,
    platforms: Sequence[str],
    target: str,
    inputs: Sequence[Task],
    group: str | None = None,
    project: Project | None = None,
) -> ManifestToolPushTask:
    project = Project.current()
    task = project.task(name, ManifestToolPushTask, group=group)
    task.template = template
    task.target = target
    task.platforms = list(platforms)
    task.depends_on(*inputs)
    return task



from __future__ import annotations

import dataclasses
from pathlib import Path

from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class PyclnTask(EnvironmentAwareDispatchTask):
    """A task to run the `pycln` formatter to either check for necessary changes or apply changes."""

    python_dependencies = ["pycln"]

    check_only: Property[bool] = Property.default(False)
    config_file: Property[Path]
    additional_args: Property[list[str]] = Property.default_factory(list)
    additional_files: Property[list[Path]] = Property.default_factory(list)

    # EnvironmentAwareDispatchTask

    def get_execute_command(self) -> list[str]:
        command = ["pycln", str(self.settings.source_directory)]
        command += self.settings.get_tests_directory_as_args()
        command += [str(p) for p in self.additional_files.get()]
        if self.check_only.get():
            command += ["--check", "--diff"]
        if self.config_file.is_filled():
            command += ["--config", str(self.config_file.get().absolute())]
        command += self.additional_args.get()
        return command

    def get_description(self) -> str:
        if self.check_only.get():
            return "Check Python imports with Pycln."
        else:
            return "Remove unused Python imports with Pycln."


@dataclasses.dataclass
class PyclnTasks:
    check: PyclnTask
    format: PyclnTask


def pycln(*, name: str = "python.pycln", project: Project | None = None) -> PyclnTasks:
    """Creates two pycln tasks, one to check and another to format. The check task will be grouped under `"lint"`
    whereas the format task will be grouped under `"fmt"`."""

    project = project or Project.current()
    check_task = project.task(f"{name}.check", PyclnTask, group="lint")
    check_task.check_only = True
    format_task = project.task(name, PyclnTask, group="fmt")
    return PyclnTasks(check_task, format_task)



""" Implements Poetry as a build system for kraken-std. """

from __future__ import annotations

import logging
import os
import shutil
import subprocess as sp
from collections.abc import Iterator, Sequence
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from kraken.common import NotSet
from kraken.common.path import is_relative_to
from kraken.common.pyenv import get_current_venv
from kraken.core import TaskStatus

from kraken.std.python.buildsystem.helpers import update_python_version_str_in_source_files
from kraken.std.python.pyproject import PackageIndex, Pyproject, PyprojectHandler
from kraken.std.python.settings import PythonSettings

from . import ManagedEnvironment, PythonBuildSystem

logger = logging.getLogger(__name__)


class PoetryPyprojectHandler(PyprojectHandler):
    """
    Pyproject configuration handler for Poetry projects.
    """

    @dataclass(frozen=True)
    class Package:
        include: str
        from_: str | None = None

    def __init__(self, pyproj: Pyproject) -> None:
        super().__init__(pyproj)

    @property
    def _poetry_section(self) -> dict[str, Any]:
        return self.raw.setdefault("tool", {}).setdefault("poetry", {})  # type: ignore[no-any-return]

    def get_packages(self, fallback: bool = True) -> list[Package]:
        """
        Returns the packages included in the distribution of this project listed in `[tool.poetry.packages]`.

        If that configuration does not exist and *fallback* is set to True, the default that Poetry will
        assume is returned.
        """

        packages: list[dict[str, Any]] | None = self._poetry_section.get("packages")
        if packages is None and fallback:
            package_name = self._poetry_section["name"]
            return [self.Package(include=package_name.replace("-", "_").replace(".", "_"))]
        else:
            return [self.Package(include=x["include"], from_=x.get("from")) for x in packages or ()]

    # PyprojectHandler

    def get_name(self) -> str:
        return self._poetry_section["name"]  # type: ignore[no-any-return]

    def get_version(self) -> str | None:
        return self._poetry_section.get("version")

    def get_python_version_constraint(self) -> str | None:
        return self._poetry_section.get("dependencies", {}).get("python")  # type: ignore[no-any-return]

    def set_version(self, version: str | None) -> None:
        project: dict[str, Any] = self._poetry_section
        if version is None:
            project.pop("version", None)
        else:
            project["version"] = version

    def get_package_indexes(self) -> list[PackageIndex]:
        sources = self._poetry_section.get("source", [])
        return [
            PackageIndex(
                alias=source["name"],
                index_url=source["url"],
                priority=PackageIndex.Priority[source["priority"]]
                if "priority" in source
                else (
                    PackageIndex.Priority.default
                    # Support deprecated source configurations.
                    if source.get("default")
                    else PackageIndex.Priority.secondary
                    if source.get("secondary")
                    else PackageIndex.Priority.supplemental
                ),
                verify_ssl=True,
            )
            for source in sources
        ]

    def set_package_indexes(self, indexes: Sequence[PackageIndex]) -> None:
        sources = self._poetry_section.setdefault("source", [])
        sources.clear()
        for index in indexes:
            if not index.verify_ssl:
                logger.warning(
                    "Poetry does not support disabling SSL verification for indexes (source: %s)", index.alias
                )
            sources.append(
                {
                    "name": index.alias,
                    "url": index.index_url,
                    "priority": index.priority.value,
                }
            )

    def set_path_dependencies_to_version(self, version: str) -> None:
        """
        Walks through the `[tool.poetry.dependencies]`, `[tool.poetry.dev-dependencies]`
        and `[tool.poetry.group.dev.dependencies]` groups to replace all path dependencies
        with proper index dependencies pointing using the specified `version` string.
        """

        def _dependency_groups() -> Iterator[tuple[str, dict[str, Any]]]:
            if dependencies := self._poetry_section.get("dependencies"):
                yield "dependencies", dependencies
            if dev_dependencies := self._poetry_section.get("dev-dependencies"):
                yield "dev-dependencies", dev_dependencies
            if group_dev_dependencies := self._poetry_section.get("group", {}).get("dev", {}).get("dependencies"):
                yield "group.dev.dependencies", group_dev_dependencies

        for name, dependencies in _dependency_groups():
            for key, value in dependencies.items():
                if isinstance(value, dict) and "path" in value:
                    logger.debug("Replacing path dependency %s with version %s in %s", key, version, name)
                    dependencies[key] = version


class PoetryPythonBuildSystem(PythonBuildSystem):
    name = "Poetry"

    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory

    # PythonBuildSystem

    def get_pyproject_reader(self, pyproject: Pyproject) -> PoetryPyprojectHandler:
        return PoetryPyprojectHandler(pyproject)

    def supports_managed_environments(self) -> bool:
        return True

    def get_managed_environment(self) -> ManagedEnvironment:
        return PoetryManagedEnvironment(self.project_directory)

    def update_lockfile(self, settings: PythonSettings, pyproject: Pyproject) -> TaskStatus:
        command = ["poetry", "update"]
        sp.check_call(command, cwd=self.project_directory)
        return TaskStatus.succeeded()

    def requires_login(self) -> bool:
        return True

    def login(self, settings: PythonSettings) -> None:
        for index in settings.package_indexes.values():
            if index.is_package_source and index.credentials:
                command = ["poetry", "config", "http-basic." + index.alias, index.credentials[0], index.credentials[1]]
                safe_command = ["poetry", "config", "http-basic." + index.alias, index.credentials[0], "MASKED"]
                logger.info("$ %s", safe_command)
                code = sp.call(command)
                if code != 0:
                    raise RuntimeError(f"command {safe_command!r} failed with exit code {code}")

    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        previous_version: str | None = None
        revert_version_paths: list[Path] = []
        if as_version is not None:
            # Bump the in-source version number.
            pyproject = self.get_pyproject_reader(Pyproject.read(self.project_directory / "pyproject.toml"))
            pyproject.set_path_dependencies_to_version(as_version)
            previous_version = pyproject.get_version()
            pyproject.set_version(as_version)
            pyproject.raw.save()
            for package in pyproject.get_packages(fallback=True):
                package_dir = self.project_directory / (package.from_ or "") / package.include
                n_replaced = update_python_version_str_in_source_files(as_version, package_dir)
                if n_replaced > 0:
                    revert_version_paths.append(package_dir)
                    print(
                        f"Bumped {n_replaced} version reference(s) in "
                        f"{package_dir.relative_to(self.project_directory)} to {as_version}"
                    )

        # Poetry does not allow configuring the output folder, so it's always going to be "dist/".
        # We remove the contents of that folder to make sure we know what was produced.
        dist_dir = self.project_directory / "dist"
        if dist_dir.exists():
            shutil.rmtree(dist_dir)

        command = ["poetry", "build"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)
        src_files = list(dist_dir.iterdir())
        dst_files = [output_directory / path.name for path in src_files]
        for src, dst in zip(src_files, dst_files):
            shutil.move(str(src), dst)

        # Unless the output directory is a subdirectory of the dist_dir, we remove the dist dir again.
        if not is_relative_to(output_directory, dist_dir):
            shutil.rmtree(dist_dir)

        # Roll back the previously updated in-source version numbers.
        if previous_version is not None:
            pyproject.set_version(previous_version)
            pyproject.raw.save()
            for package_dir in revert_version_paths:
                update_python_version_str_in_source_files(previous_version, package_dir)

        return dst_files


class PoetryManagedEnvironment(ManagedEnvironment):
    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory
        self._env_path: Path | None | NotSet = NotSet.Value

    def _get_current_poetry_environment_path(self) -> Path | None:
        """Uses `poetry env info -p`. This will not work if Poetry has to fall back to a compatible Python
        version if the version that Poetry is installed into is not compatible."""

        # Ensure we de-activate any environment that might be active when Kraken is invoked. Otherwise,
        # Poetry would fall back to that environment.
        environ = os.environ.copy()
        venv = get_current_venv(environ)
        if venv:
            venv.deactivate(environ)

        command = ["poetry", "env", "info", "-p"]
        try:
            response = sp.check_output(command, cwd=self.project_directory, env=environ).decode().strip()
        except sp.CalledProcessError as exc:
            if exc.returncode != 1:
                raise
            return None
        else:
            return Path(response)

    def _get_all_poetry_known_environment_paths(self) -> list[Path]:
        """Uses `poetry env list --full-path` to get the path to all relevant virtual environments that are known
        to Poetry for the project. We fall back to this method if `poetry env info -p` fails us."""

        command = ["poetry", "env", "list", "--full-path"]
        try:
            response = sp.check_output(command, cwd=self.project_directory).decode().strip().splitlines()
        except sp.CalledProcessError as exc:
            if exc.returncode != 1:
                raise
            return []
        else:
            return [Path(line.replace(" (Activated)", "").strip()) for line in response if line]

    def _get_poetry_environment_path(self) -> Path | None:
        # Run the two Poetry commands we need to run in parallel to improve load times.
        with ThreadPoolExecutor() as executor:
            venv_path_future = executor.submit(self._get_current_poetry_environment_path)
            known_venvs_future = executor.submit(self._get_all_poetry_known_environment_paths)
            venv_path = venv_path_future.result()
            if venv_path is not None:
                return venv_path
            known_venvs = known_venvs_future.result()
            if known_venvs:
                if len(known_venvs) > 1:
                    logger.warning(
                        "This project has multiple Poetry environments. Picking the first one (%s)", known_venvs[0]
                    )
                return known_venvs[0]
            return None

    # ManagedEnvironment

    def exists(self) -> bool:
        try:
            self.get_path()
            return True
        except RuntimeError:
            return False

    def get_path(self) -> Path:
        if self._env_path is NotSet.Value:
            self._env_path = self._get_poetry_environment_path()
        if self._env_path is None:
            raise RuntimeError("managed environment does not exist")
        return self._env_path

    def install(self, settings: PythonSettings) -> None:
        command = ["poetry", "install", "--no-interaction"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)



from importlib import import_module

import pytest

from kraken.std.docker import BUILD_BACKENDS


@pytest.mark.parametrize(argnames="backend", argvalues=BUILD_BACKENDS.keys())
def test_build_backends_importable(backend: str) -> None:
    module_name, member = BUILD_BACKENDS[backend].rpartition(".")[::2]
    module = import_module(module_name)
    print(module)
    assert hasattr(module, member)



from __future__ import annotations

import hashlib
import re
from collections.abc import Generator, Iterator
from pathlib import Path
from typing import Any, TypeVar

from kraken.core import Property, Task, TaskStatus
from kraken.core.cli.executor import status_to_text

T = TypeVar("T")
R = TypeVar("R")


class ValidateReadmeTask(Task):
    """
    This task is intended to be used to validate a README file in a project. It can validate the filename of the
    readme, check if there are multiple readme files, if the file has enough content or matches some disallowed
    content hashes.
    """

    #: The directory in which to check for the readme file.
    directory: Property[Path]

    #: The preferred name(s) of the readme file. If any file that starts with `readme` is found and not one of
    #: these files, it will issue a warning. If none of the preferred files exist, an error is issued.
    preferred_file_names: Property[list[str]] = Property.default_factory(lambda: ["README.md"])

    #: A list of disallowed content hashes. If the readme file matches any of these hashes, an warning is issued.
    #: The first line of the readme will be ignored, as well as empty lines when computing its hash. The key
    #: of the dictionary must contain a name for the hash.
    disallowed_md5_content_hashes: Property[dict[str, str]] = Property.default_factory(dict)

    #: A list of Regex patterns that should not be found in the readme file. If any of these patterns are found,
    #: a warning is issued.
    disallowed_regex_patterns: Property[list[str]] = Property.default_factory(list)

    #: The minimum number of characters that must be present in the readme file. If the readme file has less
    #: than this number of characters, a warning is issued.
    minimum_characters: Property[int] = Property.default(0)

    #: The minimum number of lines that must be present in the readme file. If the readme file has less
    #: than this number of lines, a warning is issued.
    minimum_lines: Property[int] = Property.default(0)

    #: The status of the validation.
    statuses: Property[list[TaskStatus]] = Property.output()

    @staticmethod
    def _find_readme_file(directory: Path, preferred_file_names: list[str]) -> Generator[TaskStatus, None, Path | None]:
        """
        Finds the readme file in the given directory. If no readme file is found, it will yield an error.
        """

        readme_files = [
            item
            for item in directory.iterdir()
            if item.stem.lower() == "readme" or item.stem.lower().startswith("readme.")
        ]

        preferred_readme_files = [item for item in readme_files if item.name in preferred_file_names]
        if not preferred_readme_files:
            yield TaskStatus.failed(
                f"No preferred readme file found (expected one of: {', '.join(x for x in preferred_file_names)})"
            )
            return None

        if len(readme_files) > 1:
            yield TaskStatus.warning(f"Found multiple readme files. Linting only {preferred_readme_files[0].name}")
        else:
            yield TaskStatus.succeeded(f"Found readme file: {preferred_readme_files[0].name}")

        return preferred_readme_files[0]

    @staticmethod
    def _validate_readme_contents(
        path: Path,
        disallowed_md5_content_hashes: dict[str, str],
        disallowed_regex_patterns: list[str],
        minimum_characters: int,
        minimum_lines: int,
    ) -> Iterator[TaskStatus]:
        content = path.read_text()

        if disallowed_md5_content_hashes:
            # Compute the hash of the readme file. We skip the first line; for templatized projects, this is usually
            # the name of the project, which is not relevant for the hash, and ignore empty lines.
            with open(path) as fp:
                fp.readline()
                hashable_content = "\n".join(x for x in fp.readlines() if x.lstrip())
                file_hash = hashlib.md5(hashable_content.encode("utf-8")).hexdigest()

            inverse_hashes = {v: k for k, v in disallowed_md5_content_hashes.items()}
            if file_hash in inverse_hashes:
                yield TaskStatus.warning(f"Readme file content matches disallowed hash '{inverse_hashes[file_hash]}'")
            else:
                yield TaskStatus.succeeded("Readme file content does not match any disallowed hashes")

        if disallowed_regex_patterns:
            for pattern in disallowed_regex_patterns:
                if re.search(pattern, content):
                    yield TaskStatus.warning(f"Readme file contains disallowed pattern {pattern}")
                else:
                    yield TaskStatus.succeeded(f"Readme file does not contain disallowed pattern {pattern}")

        if minimum_characters > 0:
            if len(content) < minimum_characters:
                yield TaskStatus.warning(
                    f"Readme file has less than {minimum_characters} characters (got: {len(content)})"
                )
            else:
                yield TaskStatus.succeeded(
                    f"Readme file has at least {minimum_characters} characters (got: {len(content)})"
                )

        if minimum_lines > 0:
            line_count = content.count("\n")
            if line_count + 1 < minimum_lines:
                yield TaskStatus.warning(f"Readme file has less than {minimum_lines} lines (got: {line_count + 1})")
            else:
                yield TaskStatus.succeeded(f"Readme file has at least {minimum_lines} lines (got: {line_count + 1})")

    def execute(self) -> TaskStatus:
        statuses, readme_file = collect(
            self._find_readme_file(self.directory.get_or(self.project.directory), self.preferred_file_names.get())
        )
        if readme_file is not None:
            statuses += self._validate_readme_contents(
                path=readme_file,
                disallowed_md5_content_hashes=self.disallowed_md5_content_hashes.get(),
                disallowed_regex_patterns=self.disallowed_regex_patterns.get(),
                minimum_characters=self.minimum_characters.get(),
                minimum_lines=self.minimum_lines.get(),
            )

        print(f"    Check results ({len(statuses)})")
        for status in statuses:
            print(f"      {status_to_text(status)}")

        num_checks = len(statuses)
        self.statuses.set(statuses)

        if any(x.is_failed() for x in statuses):
            return TaskStatus.failed(
                f"completed {sum(1 for x in statuses if x.is_failed())} out of {len(statuses)} check(s) with errors"
            )
        elif any(x.is_warning() for x in statuses):
            return TaskStatus.warning(
                f"completed {sum(1 for x in statuses if x.is_warning())} out of {len(statuses)} check(s) with warnings"
            )
        else:
            return TaskStatus.succeeded(f"completed {num_checks} check(s)")


def collect(gen: Generator[T, Any, R]) -> tuple[list[T], R]:
    items = []
    while True:
        try:
            items.append(next(gen))
        except StopIteration as e:
            return items, e.value



from __future__ import annotations

import shutil
import subprocess as sp
from os.path import realpath

from kraken.core import Property, Task, TaskStatus


class RustupTargetAddTask(Task):
    description = "Installs a given target for Cargo. Skipped if Cargo was installed by Nix."

    target: Property[str]

    def execute(self) -> TaskStatus:
        cargo_path = shutil.which("cargo")
        if cargo_path is not None:
            cargo_real_path = realpath(cargo_path)

            if cargo_real_path.startswith("/nix/"):
                return TaskStatus.skipped()

        command = ["rustup", "target", "add", self.target.get()]
        result = sp.call(command, cwd=self.project.directory)
        return TaskStatus.from_exit_code(command, result)



"""Manifest parser for the relevant bits and pieces."""

from __future__ import annotations

import json
import logging
import subprocess
from dataclasses import dataclass, fields
from enum import Enum
from pathlib import Path
from typing import Any

import tomli
import tomli_w
from pydantic import ClassError

logger = logging.getLogger(__name__)


@dataclass
class Bin:
    name: str
    path: str

    def to_json(self) -> dict[str, str]:
        return {"name": self.name, "path": self.path}


# TODO: Differentiate between lib kinds?


class ArtifactKind(Enum):
    BIN = 1
    LIB = 2


@dataclass
class Artifact:
    name: str
    path: str
    kind: ArtifactKind
    manifest_path: str

    def to_json(self) -> dict[str, str]:
        return {"name": self.name, "path": self.path, "kind": str(self.kind), "manifest_path": self.manifest_path}


@dataclass
class WorkspaceMember:
    id: str
    name: str
    version: str
    edition: str
    manifest_path: Path


@dataclass
class CargoMetadata:
    _path: Path
    _data: dict[str, Any]

    workspaceMembers: list[WorkspaceMember]
    artifacts: list[Artifact]
    target_directory: Path

    @classmethod
    def read(cls, project_dir: Path) -> CargoMetadata:
        cmd = [
            "cargo",
            "metadata",
            "--no-deps",
            "--format-version=1",
            "--manifest-path",
            str(project_dir / "Cargo.toml"),
        ]
        result = subprocess.run(cmd, stdout=subprocess.PIPE)
        if result.returncode != 0:
            logger.error("Stderr: %s", result.stderr)
            logger.error(f"Could not execute `{' '.join(cmd)}`, and thus can't read the cargo metadata.")
            logger.error(
                "This is due to a malformed Cargo setup, possibly due to missing registries in the config.toml."
                "See https://github.com/kraken-build/kraken-std/pull/59 for more details."
            )
            logger.error(
                "This is a known deficiency of Kraken currently - to proceed, you must comment out "
                "non-default registry imports that you have added in your Cargo.toml files, run "
                "`krakenw run :cargoSyncConfig`, and then uncomment them again."
            )
            logger.error(
                "If you continue to have this error after performing the above, you may have unrelated Cargo issues"
            )
            raise RuntimeError("Could not read cargo metadata. Please see logs for instructions.")
        else:
            return cls.of(project_dir, json.loads(result.stdout.decode("utf-8")))

    @classmethod
    def of(cls, path: Path, data: dict[str, Any]) -> CargoMetadata:
        workspace_members = []
        artifacts = []
        for package in data["packages"]:
            id = package["id"]
            if id in data["workspace_members"]:
                workspace_members.append(
                    WorkspaceMember(
                        id, package["name"], package["version"], package["edition"], Path(package["manifest_path"])
                    )
                )
                for target in package["targets"]:
                    if "bin" in target["kind"]:
                        artifacts.append(
                            Artifact(target["name"], target["src_path"], ArtifactKind.BIN, package["manifest_path"])
                        )
                    elif "lib" in target["kind"]:
                        artifacts.append(
                            Artifact(target["name"], target["src_path"], ArtifactKind.LIB, package["manifest_path"])
                        )

        return cls(path, data, workspace_members, artifacts, Path(data["target_directory"]))


@dataclass
class Package:
    name: str
    version: str | None
    edition: str | None
    unhandled: dict[str, Any] | None

    @classmethod
    def from_json(cls, json: dict[str, str]) -> Package:
        cloned = dict(json)
        name = cloned.pop("name")
        version = cloned.pop("version", None)
        edition = cloned.pop("edition", None)
        return Package(name, version, edition, cloned)

    def to_json(self) -> dict[str, str]:
        values = {f.name: getattr(self, f.name) for f in fields(self) if f.name != "unhandled"}
        if self.unhandled is not None:
            values.update({k: v for k, v in self.unhandled.items() if v is not None})
        return {k: v for k, v in values.items() if v is not None}


@dataclass
class WorkspacePackage:
    version: str
    unhandled: dict[str, Any] | None

    @classmethod
    def from_json(cls, json: dict[str, str]) -> WorkspacePackage:
        cloned = dict(json)
        version = cloned.pop("version")
        return WorkspacePackage(version, cloned)

    def to_json(self) -> dict[str, str]:
        values = {f.name: getattr(self, f.name) for f in fields(self) if f.name != "unhandled"}
        if self.unhandled is not None:
            values.update({k: v for k, v in self.unhandled.items() if v is not None})
        return {k: v for k, v in values.items() if v is not None}


@dataclass
class Workspace:
    package: WorkspacePackage | None
    members: list[str] | None
    unhandled: dict[str, Any] | None

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Workspace:
        cloned = dict(json)
        return Workspace(
            WorkspacePackage.from_json(cloned.pop("package")) if "package" in cloned else None,
            cloned.pop("members") if "members" in cloned else None,
            cloned,
        )

    def to_json(self) -> dict[str, Any]:
        values = {
            "package": self.package.to_json() if self.package else None,
            "members": self.members if self.members else None,
        }
        if self.unhandled is not None:
            values.update({k: v for k, v in self.unhandled.items() if v is not None})
        return {k: v for k, v in values.items() if v is not None}


@dataclass
class Dependencies:
    data: dict[str, Any]

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Dependencies:
        cloned = dict(json)
        return Dependencies(cloned)

    def to_json(self) -> dict[str, Any]:
        return self.data


@dataclass
class CargoManifest:
    _path: Path
    _data: dict[str, Any]

    package: Package | None
    workspace: Workspace | None
    dependencies: Dependencies | None
    build_dependencies: Dependencies | None
    bin: list[Bin]

    @classmethod
    def read(cls, path: Path) -> CargoManifest:
        with path.open("rb") as fp:
            ret = cls.of(path, tomli.load(fp))
            if ret.package is None and ret.workspace is None:
                raise ClassError
            return ret

    @classmethod
    def of(cls, path: Path, data: dict[str, Any]) -> CargoManifest:
        return cls(
            path,
            data,
            Package.from_json(data["package"]) if "package" in data else None,
            Workspace.from_json(data["workspace"]) if "workspace" in data else None,
            Dependencies.from_json(data["dependencies"]) if "dependencies" in data else None,
            Dependencies.from_json(data["build-dependencies"]) if "build-dependencies" in data else None,
            [Bin(**x) for x in data.get("bin", [])],
        )

    def to_json(self) -> dict[str, Any]:
        result = self._data.copy()
        if self.bin:
            result["bin"] = [x.to_json() for x in self.bin]
        else:
            result.pop("bin", None)
        if self.package:
            result["package"] = self.package.to_json()
        if self.workspace:
            result["workspace"] = self.workspace.to_json()
        if self.dependencies:
            result["dependencies"] = self.dependencies.to_json()
        if self.build_dependencies:
            result["build-dependencies"] = self.build_dependencies.to_json()
        return result

    def to_toml_string(self) -> str:
        return tomli_w.dumps(self.to_json())

    def save(self, path: Path | None = None) -> None:
        path = path or self._path
        with path.open("wb") as fp:
            tomli_w.dump(self.to_json(), fp)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from kraken.core import Project, VoidTask


@dataclass
class Resource:
    """A resource represents a file or directory on the file system."""

    name: str
    path: Path


@dataclass
class BinaryArtifact(Resource):
    """A subclass of a resource that represents a binary artifact."""


@dataclass
class LibraryArtifact(Resource):
    """A subclass of a resource that represents a library artifact."""


def resource(*, name: str, path: str | Path, project: Project | None = None) -> Resource:
    """Creates a task for the purpose of carrying a :class:`Resource` descriptor."""

    project = project or Project.current()
    resource = Resource(name, project.directory / path)
    task = project.task(name, VoidTask)
    task.outputs.append(resource)
    return resource



from __future__ import annotations

import os
import shlex
import subprocess as sp
import time
from dataclasses import dataclass

from kraken.core import Project, Property, Task, TaskStatus

from kraken.std.cargo.manifest import ArtifactKind, CargoMetadata
from kraken.std.descriptors.resource import BinaryArtifact, LibraryArtifact


@dataclass
class CargoBinaryArtifact(BinaryArtifact):
    pass


@dataclass
class CargoLibraryArtifact(LibraryArtifact):
    pass


class CargoBuildTask(Task):
    """This task runs `cargo build` using the specified parameters. It will respect the authentication
    credentials configured in :attr:`CargoProjectSettings.auth`."""

    #: The build target (debug or release). If this is anything else, the :attr:`out_binaries` will be set
    #: to an empty list instead of parsed from the Cargo manifest.
    target: Property[str]

    #: Additional arguments to pass to the Cargo command-line.
    additional_args: Property[list[str]] = Property.default_factory(list)

    #: Whether to build incrementally or not.
    incremental: Property[bool | None] = Property.default(None)

    #: Environment variables for the Cargo command.
    env: Property[dict[str, str]] = Property.default_factory(dict)

    #: Number of times to retry before failing this job
    retry_attempts: Property[int] = Property.default(0)

    #: An output property for the Cargo binaries that are being produced by this build.
    out_binaries: Property[list[CargoBinaryArtifact]] = Property.output()

    #: An output property for the Cargo libraries that are being produced by this build.
    out_libraries: Property[list[CargoLibraryArtifact]] = Property.output()

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)

    def get_description(self) -> str | None:
        command = self.get_cargo_command({})
        self.make_safe(command, {})
        return f"Run `{' '.join(command)}`."

    def get_cargo_command_additional_flags(self) -> list[str]:
        return shlex.split(os.environ.get("KRAKEN_CARGO_BUILD_FLAGS", ""))

    def get_cargo_command(self, env: dict[str, str]) -> list[str]:
        incremental = self.incremental.get()
        if incremental is not None:
            env["CARGO_INCREMENTAL"] = "1" if incremental else "0"
        return ["cargo", "build"] + self.additional_args.get()

    def make_safe(self, args: list[str], env: dict[str, str]) -> None:
        pass

    def execute(self) -> TaskStatus:
        env = self.env.get()
        command = self.get_cargo_command(env) + self.get_cargo_command_additional_flags()

        safe_command = command[:]
        safe_env = env.copy()
        self.make_safe(safe_command, safe_env)
        self.logger.info("%s [env: %s]", safe_command, safe_env)

        out_binaries: list[CargoBinaryArtifact] = []
        out_libraries_candidates: list[CargoLibraryArtifact] = []
        if self.target.get_or(None) in ("debug", "release"):
            # Expose the output binaries that are produced by this task.
            # We only expect a binary to be built if the target is debug or release.
            manifest = CargoMetadata.read(self.project.directory)
            target_dir = manifest.target_directory / self.target.get()
            for artifact in manifest.artifacts:
                # Rust binaries have an extensionless name whereas libraries are prefixed with "lib" and suffixed with
                #
                # - ".rlib" for Rust libraries
                # - ".so" (Linux), ".dylib" (macOS) or ".dll" (Windows) for dynamic Rust and system libraries
                # - ".a" (Linux, macOS) or ".lib" (Windows) for static system libraries
                if artifact.kind is ArtifactKind.BIN:
                    out_binaries.append(CargoBinaryArtifact(artifact.name, target_dir / artifact.name))
                elif artifact.kind is ArtifactKind.LIB:
                    base_name = f"lib{artifact.name}"
                    for file_extension in ["rlib", "so", "dylib", "dll", "a", "lib"]:
                        filename = ".".join([base_name.replace("-", "_"), file_extension])
                        out_libraries_candidates.append(CargoLibraryArtifact(base_name, target_dir / filename))

        total_attempts = self.retry_attempts.get() + 1

        while total_attempts > 0:
            result = sp.call(command, cwd=self.project.directory, env={**os.environ, **env})

            if result == 0:
                # Check that binaries which were due have been built.
                for out_bin in out_binaries:
                    assert out_bin.path.is_file(), out_bin
                self.out_binaries.set(out_binaries)

                # Check that at least one library has been built if libraries were due.
                out_libraries = []
                if len(out_libraries_candidates) != 0:
                    for out_libraries_candidate in out_libraries_candidates:
                        # Since we generate all possible file extensions, we must only keep the ones that exist
                        if out_libraries_candidate.path.is_file():
                            out_libraries.append(out_libraries_candidate)
                    assert (
                        len(out_libraries) != 0
                    ), f'No libraries were built even though some were due, e.g. "{out_libraries_candidates[0].name}"'
                self.out_libraries.set(out_libraries)
                break
            else:
                total_attempts -= 1
                self.logger.warn("%s failed with result %s", safe_command, result)
                self.logger.warn("There are %s attempts remaining", total_attempts)
                if total_attempts > 0:
                    self.logger.info("Waiting for 10 seconds before retrying..")
                    time.sleep(10)

        return TaskStatus.from_exit_code(safe_command, result)



"""
A simple parser for `.gitignore` files which keeps track of sections of comments.
"""

from __future__ import annotations

from collections.abc import Iterable
from io import StringIO
from itertools import islice
from os import PathLike
from pathlib import Path
from typing import Literal, NamedTuple


class GitignoreEntry(str):
    @property
    def type(self) -> Literal["comment", "blank", "path"]:
        if not self:
            return "blank"
        if self.startswith("#"):
            return "comment"
        return "path"


class GitignoreFile(list[GitignoreEntry]):
    """
    Represents a `.gitignore` file, keeping track of the entries and the generated content.
    """

    def find_comment(self, comment: str) -> int | None:
        return next((i for i, e in enumerate(self) if e.type == "comment" and e.lstrip("#").strip() == comment), None)

    def paths(self, start: int | None = None, stop: int | None = None) -> Iterable[str]:
        return (entry for entry in islice(self, start, stop) if entry.type == "path")

    def add_comment(self, comment: str, index: int | None = None) -> None:
        assert "\n" not in comment
        entry = GitignoreEntry("# " + comment)
        self.insert(len(self) if index is None else index, entry)

    def add_blank(self, index: int | None = None) -> None:
        entry = GitignoreEntry()
        self.insert(len(self) if index is None else index, entry)

    def add_path(self, path: str, index: int | None = None) -> None:
        assert "\n" not in path
        entry = GitignoreEntry(path)
        assert entry.type == "path", entry
        self.insert(len(self) if index is None else index, entry)

    def remove_path(self, path: str) -> None:
        removed = 0
        while True:
            index = next((i for i, e in enumerate(self) if e.type == "path" and e == path), None)
            if index is None:
                break
            del self[index]
            removed += 1
        if removed == 0:
            raise ValueError(f'"{path}" not in GitignoreFile')

    def sort_gitignore(self, sort_paths: bool = True, sort_groups: bool = False) -> None:
        """
        Sorts the entries in the gitignore file, keeping paths under a common comment block grouped.
        Will also get rid of any extra blanks.

        :param gitignore: The input to sort.
        :param sort_paths: Whether to sort paths (default: True).
        :param sort_groups: Whether to sort groups among themselves, not just paths within groups (default: False).
        :return: A new, sorted gitignore file.
        """

        class Group(NamedTuple):
            comments: list[str]
            paths: list[str]

        # List of (comments, paths).
        groups: list[Group] = [Group([], [])]

        for entry in self:
            if entry.type == "path":
                groups[-1].paths.append(entry)
            elif entry.type == "comment":
                # If we already have paths in the current group, we open a new group.
                if groups[-1].paths:
                    groups.append(Group([entry], []))
                # Otherwise we append the comment to the group.
                else:
                    groups[-1].comments.append(entry)

        if sort_groups:
            groups.sort(key=lambda g: "\n".join(g.comments).lower())

        self.clear()
        self.add_blank()  # separate GENERATED from USER content
        for group in groups:
            if sort_paths:
                group.paths.sort(key=str.lower)
            for comment in group.comments:
                self.add_comment(comment)
            for path in group.paths:
                self.add_path(path)
            self.add_blank()

        if self and self[-1].type == "blank":
            self.pop()

    @staticmethod
    def parse(file: Path | str | Iterable[str]) -> GitignoreFile:
        match file:
            case str():
                return GitignoreFile.parse(StringIO(file))
            case PathLike():
                with file.open() as fp:
                    return GitignoreFile.parse(fp)

        gitignore = GitignoreFile()
        for line in file:
            line = line.rstrip("\n")
            if line.strip():
                gitignore.append(GitignoreEntry(line))
            else:
                gitignore.append(GitignoreEntry())
        return gitignore



from __future__ import annotations

import os
import subprocess as sp
from pathlib import Path

from kraken.core import Property, Task, TaskStatus


class CargoBaseSqlxTask(Task):
    base_directory: Property[Path]
    database_url: Property[str]

    def _execute_command(self, arguments: list[str]) -> TaskStatus:
        command = ["cargo", "sqlx", *arguments]
        if self.database_url.is_filled():
            command.extend(["--database-url", self.database_url.get()])

        base_directory = self.base_directory.get() if self.base_directory.is_filled() else self.project.directory

        result = sp.call(command, cwd=base_directory, env={**os.environ})

        return TaskStatus.from_exit_code(command, result)



import textwrap

from kraken.core import BuildError, Project
from pytest import raises

from kraken.std.git import gitignore
from kraken.std.git.tasks.sync_task import GitignoreSyncTask

NO_GENERATED_CONTENT = """
# This is a comment
and/a/path
"""

WITH_GENERATED_CONTENT = """
foo
# GENERATED-CONTENT-START
bar
# GENERATED-CONTENT-END
baz
"""


def test__GitignoreSyncTask__inserts_generated_content_on_top(kraken_project: Project) -> None:
    path = kraken_project.directory / ".gitignore"
    path.write_text(NO_GENERATED_CONTENT)
    task = kraken_project.task("gitignore", GitignoreSyncTask)
    task.finalize()
    print(task.prepare())
    print(task.execute())
    assert (
        path.read_text()
        == textwrap.dedent(
            """
        # GENERATED-CONTENT-START
        # Kraken
        /build
        # GENERATED-CONTENT-END

        # This is a comment
        and/a/path
        """
        ).lstrip()
    )


def test__GitignoreSyncTask__inserts_generated_content_on_bottom(kraken_project: Project) -> None:
    path = kraken_project.directory / ".gitignore"
    path.write_text(NO_GENERATED_CONTENT)
    task = kraken_project.task("gitignore", GitignoreSyncTask)
    task.where = "bottom"
    task.finalize()
    print(task.prepare())
    print(task.execute())
    assert path.read_text() == textwrap.dedent(
        """
        # This is a comment
        and/a/path

        # GENERATED-CONTENT-START
        # Kraken
        /build
        # GENERATED-CONTENT-END
        """
    )


def test__GitignoreSyncTask__updates_existing_generated_content(kraken_project: Project) -> None:
    path = kraken_project.directory / ".gitignore"
    path.write_text(WITH_GENERATED_CONTENT)
    task = kraken_project.task("gitignore", GitignoreSyncTask)
    task.finalize()
    print(task.prepare())
    print(task.execute())
    assert path.read_text() == textwrap.dedent(
        """
        foo

        # GENERATED-CONTENT-START
        # Kraken
        /build
        # GENERATED-CONTENT-END

        baz
        """
    )


def test__gitignore__check_and_apply_tasks(kraken_project: Project) -> None:
    gitignore(gitignore_io_tokens=["python"])[0]
    with raises(BuildError):
        kraken_project.context.execute([":check"])
    kraken_project.context.execute([":apply"])
    kraken_project.context.execute([":check"])



from __future__ import annotations

import logging
import re
import subprocess as sp

from kraken.core import Property, Task, TaskStatus


class CargoCheckToolchainVersionTask(Task):
    description = "Ensures that Cargo (and so, Rust) is at least at the given version"

    minimal_version: Property[str]

    def execute(self) -> TaskStatus:
        minimal_version = self.minimal_version.get()

        # we fetch Cargo version
        try:
            result = sp.check_output(["cargo", "--verbose", "--version"]).decode()
            cargo_metadata = {key: value for key, value in re.findall(r"^([a-z]+):\s+(.*)\s*$", result, re.MULTILINE)}
        except sp.CalledProcessError as e:
            logging.error(f"Rust Cargo tool returned error code {e.returncode}, are you sure it is properly installed?")
            logging.info("You can install cargo using https://rustup.rs/ or `brew install rustup-init` on macOS")
            return TaskStatus.failed("cargo not found")
        if "release" not in cargo_metadata:
            return TaskStatus.failed("No release found in cargo metadata")
        cargo_version = cargo_metadata["release"]

        try:
            parsed_cargo_version = self._parse_version(cargo_version)
        except ValueError:
            return TaskStatus.failed(f"Invalid cargo version: {cargo_version}")
        try:
            parsed_minimal_version = self._parse_version(minimal_version)
        except ValueError:
            return TaskStatus.failed(f"Invalid expected version: {minimal_version}")

        # Hack: the order on Python tuples is the lexicographic order, the one we want here
        if parsed_minimal_version > parsed_cargo_version:
            logging.warning(
                f"The cargo version is outdated. Expecting at least {minimal_version} but found {cargo_version}"
            )
            logging.info("Please upgrade cargo using `rustup update`")
            return TaskStatus.failed(f"Outdated version {cargo_version}")

        return TaskStatus.succeeded()

    @staticmethod
    def _parse_version(version: str) -> tuple[int, ...]:
        return tuple(int(e) for e in version.split("-")[0].split("."))



from __future__ import annotations

import os
import subprocess as sp
import tempfile
from pathlib import Path

from kraken.common import flatten, not_none
from kraken.core import Project, Property, TaskStatus

from kraken.std.docker.util.dockerfile import update_run_commands

from .base_build_task import BaseBuildTask


class DockerBuildTask(BaseBuildTask):
    """Implements building a Docker image using the native `docker build` command."""

    #: Whether to use Docker Buildkit. Enabled by default.
    native_use_buildkit: Property[bool] = Property.default(True)

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.preprocess_dockerfile.set(True)

    # BaseBuildTask overrides

    def _preprocess_dockerfile(self, dockerfile: Path) -> str:
        mount_string = " ".join(f"--mount=type=secret,id={sec}" for sec in self.secrets.get().keys()) + " "
        return update_run_commands(dockerfile.read_text(), prefix=mount_string)

    # Task overrides

    def finalize(self) -> None:
        if self.cache_repo.get() is not None:
            self.logger.warning("cache_repo is not supported for DockerBuildTask")
        if self.squash.get() is not None:
            self.logger.warning("squash is not supported for DockerBuildTask")
        if not self.load.get():
            self.logger.warning("load is not supported for DockerBuildTask, resulting image will always be loaded")
        if self.push.get() and not self.tags.get():
            raise ValueError(f"{self}.tags cannot be empty if .push is enabled")
        return super().finalize()

    def execute(self) -> TaskStatus:
        command = ["docker", "build", str(self.build_context.get().absolute())]
        if self.dockerfile.is_filled():
            command += ["-f", str(self.dockerfile.get().absolute())]
        if self.platform.is_filled():
            command += ["--platform", str(self.platform.get())]
        command += flatten(["--build-arg", f"{k}={v}"] for k, v in self.build_args.get().items())
        if self.cache_repo.get():
            # NOTE (@NiklasRosenstein): Buildx does not allow leading underscores, while Kaniko and Artifactory do.
            command += ["--cache-from", f"type=registry,ref={not_none(self.cache_repo.get())}"]
        if not self.cache.get():
            command += ["--no-cache"]
        command += flatten(["--tag", t] for t in self.tags.get())
        if self.target.get():
            command += ["--target", not_none(self.target.get())]
        if self.image_output_file.get():
            command += ["--output", f"type=tar,dest={self.image_output_file.get()}"]

        command += ["--pull", "--progress", "plain"]

        # Buildx will take the secret from the environment variasbles.
        env = os.environ.copy()
        env["DOCKER_BUILDKIT"] = "1" if self.native_use_buildkit.get() else "0"

        # TODO (@nrosenstein): docker login for auth

        with tempfile.TemporaryDirectory() as tempdir:
            for key, value in self.secrets.get().items():
                secret_file = Path(tempdir) / key
                secret_file.write_text(value)
                command += ["--secret", f"id={key},src={secret_file}"]

            self.logger.info("%s", command)
            result = sp.call(command, env=env, cwd=self.project.directory)
            if result != 0:
                return TaskStatus.from_exit_code(command, result)

        if self.push.get():
            command = ["docker", "push"] + self.tags.get()
            self.logger.info("%s", command)
            result = sp.call(command, env=env, cwd=self.project.directory)

        return TaskStatus.succeeded()
